#pragma once
#include "D3DGraphics.h"
#include "TexturedQuad.h"
#include <random>

class Bubble
{
public:
	class Drawable : public ::Drawable
	{
	public:
		Drawable( const Bubble& parent )
			:
			parent( parent )
		{}
		virtual void Rasterize( D3DGraphics& gfx ) const override
		{
			const float wiggle = parent.GetWiggle();
			auto d = parent.bubbleTexture.GetDrawableAlpha();
			d.Transform(
				trans *
				Mat3::Translation( parent.locus + parent.pos ) *
				Mat3::Stretching( 1.0f + wiggle,1.0f - wiggle ) );
			d.Clip( clip );
			d.Rasterize( gfx );

			const Vec2 center = trans * ( parent.locus + parent.pos );
			gfx.DrawString( parent.text,RectF( center,50.0f ),
				parent.font,parent.textColor );
		}
	private:
		const Bubble& parent;
	};
public:
	Bubble( const std::wstring& text,Vec2 locus,const TexturedQuad& tex,const Font& font )
		:
		text( text ),
		locus( locus ),
		bubbleTexture( tex ),
		font( font ),
		distributionRot( 0.0f,deviationRot )
	{
		thetaPerterb = std::uniform_real_distribution<float>( 0.0f,2.0f * PI )( gen );
		wiggleTime = std::uniform_real_distribution<float>( 0.0f,1.0f / frequencyWiggle )( gen );
	}
	void Update( float dt )
	{
		wiggleTime += dt;
		thetaPerterb += distributionRot( gen ) * dt;
		const float len = pos.LenSq();
		const Vec2 restoreForce = len == 0.0f ? Vec2{ 0.0f,0.0f } :
			-pos * kRestore;
		vel += ( ( restoreForce +
			Vec2 { 1.0f,0.0f }.Rotation( thetaPerterb ) * forcePerterb ) / mass ) * dt;
		pos += vel;
	}
	Drawable GetDrawable() const
	{
		return Drawable( *this );
	}
private:
	float GetWiggle() const
	{
		return amplitudeWiggle * sin( 2.0f * PI * frequencyWiggle * wiggleTime );
	}
public:
	static void InitRNG()
	{
		std::random_device rd;
		gen.seed( rd() );
	}
private:
	Vec2 locus;
	Vec2 pos = { 0.0f,0.0f };
	Vec2 vel = { 0.0f,0.0f };
	float thetaPerterb;
	float wiggleTime;
	const std::wstring text;
	const float mass = 4.0f;
	const float kRestore = 0.2f;
	const float forcePerterb = 2.0f;
	const float deviationRot = PI / 1.5f;
	const float amplitudeWiggle = 0.1f;
	const float frequencyWiggle = 0.4f;
	const Color textColor = YELLOW;
	const TexturedQuad& bubbleTexture;
	const Font& font;
	std::normal_distribution<float> distributionRot;
	static std::mt19937 gen;
};